//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "camera.h"

Camera::Camera(const shared_t &glbShare, const QString &confFile, const QString &statDir, RamDisk *ram, ProcControl *proc, QCoreApplication *parent) : QObject(nullptr)
{
    Q_UNUSED(parent);

    fsW       = new QFileSystemWatcher(this);
    statTimer = new QTimer(this);
    shared    = glbShare;
    procCon   = proc;
    statPath  = statDir;
    ramDisk   = ram;
    evtLoop   = nullptr;
    detLoop   = nullptr;
    recLoop   = nullptr;
    delOnZero = false;
    objCount  = 0;

    statTimer->setInterval(5000);
    statTimer->start();

    connect(fsW,       &QFileSystemWatcher::fileChanged, this, &Camera::confChanged);
    connect(proc,      &ProcControl::prepForClose,       this, &Camera::prepForDel);
    connect(statTimer, &QTimer::timeout,                 this, &Camera::updateStat);

    start(confFile);
}

void Camera::objMinusOne()
{
    objCount--;

    if (objCount == 0)
    {
        thr1->deleteLater();
        thr2->deleteLater();
        thr3->deleteLater();

        if (delOnZero)
        {
            procCon->objMinusOne(shared.conf);

            QDir(shared.buffPath).removeRecursively();

            deleteLater();
        }
    }
}

void Camera::prepForDel()
{
    ramDisk->rmCam(shared.camName);
    statTimer->blockSignals(true);

    delOnZero = true;

    emit stop();
}

void Camera::updateStat()
{
    auto statFile = statPath + "/" + shared.camName;

    QFile file(statFile);

    file.open(QFile::WriteOnly);
    file.write(statusLine().toUtf8());
    file.close();
}

void Camera::start(const QString &conf)
{
    if (!rdConf(conf, &shared))
    {
        deleteLater();
    }
    else
    {
        ramDisk->mkdir(QString("/" + shared.camName).toUtf8().data(), 0);
        ramDisk->mkdir(QString("/" + shared.camName + "/vid").toUtf8().data(), 0);
        ramDisk->mkdir(QString("/" + shared.camName + "/img").toUtf8().data(), 0);

        ramDisk->addCam(shared.camName);
        procCon->objPlusOne(shared.conf);

        if (!fsW->files().contains(conf)) fsW->addPath(conf);

        thr1 = new QThread(nullptr);
        thr2 = new QThread(nullptr);
        thr3 = new QThread(nullptr);

        evtLoop = new EventLoop(&shared,  thr1, nullptr);
        detLoop = new DetectLoop(&shared, thr2, nullptr);
        recLoop = new RecordLoop(&shared, thr3, nullptr);

        connect(this, &Camera::stop, thr1, &QThread::quit);
        connect(this, &Camera::stop, thr2, &QThread::quit);
        connect(this, &Camera::stop, thr3, &QThread::quit);

        connect(thr1, &QThread::finished, this, &Camera::objMinusOne);
        connect(thr2, &QThread::finished, this, &Camera::objMinusOne);
        connect(thr3, &QThread::finished, this, &Camera::objMinusOne);

        connect(detLoop, &DetectLoop::starving,        recLoop, &RecordLoop::restart);
        connect(ramDisk, &RamDisk::watchedFileFlushed, detLoop, &DetectLoop::fileFlushed);

        thr1->start();
        thr2->start();
        thr3->start();

        objCount = 3;

        auto global = QDir::cleanPath(shared.confPath) + "/global";

        if (QFileInfo::exists(global))
        {
            rdConf(global, &shared, false);
        }
    }
}

void Camera::confChanged(const QString &path)
{
    Q_UNUSED(path);

    prepForDel();
}

QString Camera::statusLine()
{
    return "Detection: "       + detLoop->statusLine() +
           " Event-Scraping: " + evtLoop->statusLine() +
           " Recording: "      + recLoop->statusLine() +
           " Name: "           + shared.camName + "\n";
}
