//    This file is part of JustMotion.

//    JustMotion is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.

//    JustMotion is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.

#include "detect_loop.h"

DetectLoop::DetectLoop(shared_t *sharedRes, QThread *thr, QObject *parent) : QObject(parent)
{
    pcTimer = 0;
    shared  = sharedRes;
    starved = true;

    connect(thr, &QThread::started,  this, &DetectLoop::init);
    connect(thr, &QThread::finished, this, &DetectLoop::deleteLater);

    moveToThread(thr);
}

void DetectLoop::init()
{
    pcTimer    = new QTimer(this);
    checkTimer = new QTimer(this);

    connect(pcTimer,    &QTimer::timeout, this, &DetectLoop::pcBreak);
    connect(checkTimer, &QTimer::timeout, this, &DetectLoop::eTimeOut);

    pcTimer->start(shared->postSecs * 1000);
    checkTimer->start(5000);
    checkTimer->setSingleShot(true);
}

void DetectLoop::reset()
{
    eventQue.inQue  = false;
    eventQue.score  = 0;
    eventQue.queAge = 0;

    eventQue.imgPath.clear();
    eventQue.vidList.clear();
    eventQue.timeStamp.clear();
}

void DetectLoop::eTimeOut()
{
    starved = true;

    emit starving();
}

void DetectLoop::pcBreak()
{
    if (!shared->postCmd.isEmpty())
    {
        checkTimer->stop();

        qInfo() << "---POST_BREAK---";

        if (eventQue.inQue)
        {
            qInfo() << "motion detected, skipping the post command";
        }
        else
        {
            qInfo() << "no motion detected, running post command: " << shared->postCmd;

            auto args = parseArgs(shared->postCmd.toUtf8(), -1);

            if (args.isEmpty())
            {
                qCritical() << "err: did not parse an executable from the post command line.";
            }
            else
            {
                QProcess::execute(args[0], args.mid(1));
            }
        }

        checkTimer->start();
    }
}

float DetectLoop::getFloatFromExe(const QByteArray &line)
{
    QString strLine(line);
    QString strNum;

    for (auto chr : strLine)
    {
        if (chr.isDigit() || (chr == '.'))
        {
            strNum.append(chr);
        }
        else
        {
            break;
        }
    }

    auto ok  = false;
    auto res = strNum.toFloat(&ok);

    if (!ok || strNum.isEmpty())
    {
        qCritical() << "err: the image comp command returned unexpected output and couldn't be converted to float";
        qCritical() << "     raw output: " << line;
    }

    return res;
}

QStringList DetectLoop::buildArgs(const QString &prev, const QString &next)
{
    auto args = parseArgs(shared->compCmd.toUtf8(), -1);

    for (auto i = 0; i < args.size(); ++i)
    {
        if (args[i] == PREV_IMG) args[i] = prev;
        if (args[i] == NEXT_IMG) args[i] = next;
    }

    return args;
}

QStringList DetectLoop::buildSnapArgs(const QString &vidSrc, const QString &imgPath)
{
    QStringList ret;

    ret.append("-hide_banner");
    ret.append("-loglevel");
    ret.append("panic");
    ret.append("-y");
    ret.append("-i");
    ret.append(vidSrc);
    ret.append("-frames:v");
    ret.append("1");
    ret.append(imgPath);

    return ret;
}

QString DetectLoop::statusLine()
{
    if (starved)
    {
        return "STARVED";
    }
    else
    {
        return "OK     ";
    }
}

bool DetectLoop::pathFilter(const QString &path, const QString &subName)
{
    return path.startsWith("/" + shared->camName + "/" + subName) && path.endsWith(shared->streamExt);
}

void DetectLoop::fileFlushed(const QString &path)
{
    if (pathFilter(path, "vid"))
    {
        if (vidAPath.isEmpty())
        {
            vidAPath = shared->varPath + "/stream" + path;
            imgAPath = QDir::cleanPath(shared->buffPath) + QDir::separator() + QString("img") + QDir::separator() + QFileInfo(vidAPath).baseName() + ".bmp";

            QProcess::execute("ffmpeg", buildSnapArgs(vidAPath, imgAPath));
        }
        else
        {
            auto vidBPath = shared->varPath + "/stream" + path;
            auto imgBPath = QDir::cleanPath(shared->buffPath) + QDir::separator() + QString("img") + QDir::separator() + QFileInfo(vidBPath).baseName() + ".bmp";

            QProcess::execute("ffmpeg", buildSnapArgs(vidBPath, imgBPath));

            starved = false;

            checkTimer->start();

            auto compArgs = buildArgs(imgAPath, imgBPath);

            if (compArgs.isEmpty())
            {
                qCritical() << "err: could not parse a executable name from img_comp_cmd: " << shared->compCmd;
            }
            else
            {
                QProcess extComp;

                extComp.start(compArgs[0], compArgs.mid(1));
                extComp.waitForFinished();

                float score = 0;

                if (shared->outputType == "stdout")
                {
                    score = getFloatFromExe(extComp.readAllStandardOutput());
                }
                else
                {
                    score = getFloatFromExe(extComp.readAllStandardError());
                }

                qInfo() << compArgs.join(" ") << " --result: " << QString::number(score);

                if (eventQue.inQue)
                {
                    eventQue.queAge += 2;

                    if (eventQue.score < score)
                    {
                        eventQue.score   = score;
                        eventQue.imgPath = imgBPath;
                    }

                    eventQue.vidList.append(vidBPath);

                    if (eventQue.queAge >= shared->evMaxSecs)
                    {
                        shared->recList.append(eventQue);

                        reset();
                    }
                }
                else if (score >= shared->imgThresh)
                {
                    qInfo() << "--threshold_meet: " << QString::number(shared->imgThresh);

                    eventQue.score     = score;
                    eventQue.imgPath   = imgBPath;
                    eventQue.inQue     = true;
                    eventQue.queAge    = 0;
                    eventQue.timeStamp = QDateTime::currentDateTime().toString(DATETIME_FMT);

                    eventQue.vidList.append(vidAPath);
                    eventQue.vidList.append(vidBPath);
                }
            }

            vidAPath = vidBPath;
            imgAPath = imgBPath;
        }
    }
}
